@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2009-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2009-2022 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -8,7 +8,8 @@
 
 """ client_launcher.py
 
-This is a simple GUI for starting the xpra client.
+This was meant to be a simple GUI for starting the xpra client.
+Then it morphed into something quite messy. Sorry.
 
 """
 
@@ -17,39 +18,48 @@ import sys
 import traceback
 
 import gi
-gi.require_version("Gtk", "3.0")
-gi.require_version("Pango", "1.0")
-gi.require_version("GdkPixbuf", "2.0")
+gi.require_version("Gtk", "3.0")  # @UndefinedVariable
+gi.require_version("Pango", "1.0")  # @UndefinedVariable
+gi.require_version("GdkPixbuf", "2.0")  # @UndefinedVariable
+# @UnresolvedImport pylint: disable=wrong-import-position
 from gi.repository import Pango, GLib, Gtk, GdkPixbuf
 
 from xpra.gtk_common.gobject_compat import register_os_signals
 from xpra.scripts.config import read_config, make_defaults_struct, validate_config, save_config
-from xpra.codecs.codec_constants import PREFERRED_ENCODING_ORDER
 from xpra.gtk_common.gtk_util import (
     add_close_accel, scaled_image, color_parse,
     choose_file, imagebutton, get_icon_pixbuf,
     )
 from xpra.util import DEFAULT_PORT, csv, repr_ellipsized
-from xpra.os_util import WIN32, OSX, POSIX
+from xpra.os_util import WIN32, OSX
 from xpra.make_thread import start_thread
-from xpra.client.gtk_base.menu_helper import make_encodingsmenu
 from xpra.gtk_common.about import about
 from xpra.scripts.main import (
-    connect_to, make_client, is_local,
-    add_ssh_args, parse_ssh_string, add_ssh_proxy_args,
+    connect_to, make_client,
     configure_network, configure_env, configure_logging,
     )
-from xpra.platform import get_username
+from xpra.scripts.parsing import is_local, add_ssh_args, parse_ssh_option, add_ssh_proxy_args
+from xpra.exit_codes import RETRY_EXIT_CODES, EXIT_STR
+from xpra.platform.info import get_username
 from xpra.log import Logger, enable_debug_for
 
 log = Logger("launcher")
 
+MODE_NESTED_SSH = "ssh -> ssh"
+MODE_SSH = "ssh"
+MODE_TCP = "tcp"
+MODE_SSL = "ssl"
+MODE_WS = "ws"
+MODE_WSS = "wss"
+
+
 #what we save in the config file:
 SAVED_FIELDS = [
     "username", "password", "host", "port", "mode", "ssh_port",
     "encoding", "quality", "min-quality", "speed", "min-speed",
     "proxy_port", "proxy_username", "proxy_key", "proxy_password",
-    "proxy_host"
+    "proxy_host",
+    "autoconnect",
     ]
 
 #options not normally found in xpra config file
@@ -73,7 +83,7 @@ LAUNCHER_DEFAULTS = {
                         "port"              : -1,
                         "username"          : get_username(),
                         "password"          : "",
-                        "mode"              : "tcp",    #tcp,ssh,..
+                        "mode"              : MODE_TCP,    #tcp,ssh,..
                         "autoconnect"       : False,
                         "ssh_port"          : 22,
                         "proxy_host"        : "",
@@ -89,6 +99,7 @@ red     = color_parse("red")
 white   = color_parse("white")
 
 
+
 def get_active_item_index(optionmenu):
     menu = optionmenu.get_popup()
     for i, x in enumerate(menu.get_children()):
@@ -106,6 +117,7 @@ def set_history_from_active(optionmenu):
 
 def has_mdns():
     try:
+        # pylint: disable=import-outside-toplevel
         from xpra.net.mdns import get_listener_class
         lc = get_listener_class()
         log("mdns listener class: %s", lc)
@@ -119,11 +131,16 @@ def noop(*args):
     log("noop%s", args)
 
 
+# pylint: disable=import-outside-toplevel
+
+
 class ApplicationWindow:
 
     def __init__(self):
         # Default connection options
-        self.config = make_defaults_struct(extras_defaults=LAUNCHER_DEFAULTS, extras_types=LAUNCHER_OPTION_TYPES, extras_validation=self.get_launcher_validation())
+        self.config = make_defaults_struct(extras_defaults=LAUNCHER_DEFAULTS,
+                                           extras_types=LAUNCHER_OPTION_TYPES,
+                                           extras_validation=self.get_launcher_validation())
         self.parse_ssh()
         #TODO: the fixup does not belong here?
         from xpra.scripts.main import fixup_options
@@ -136,24 +153,29 @@ class ApplicationWindow:
         self.current_error = None
 
     def parse_ssh(self):
-        ssh_cmd = parse_ssh_string(self.config.ssh)[0].strip().lower()
+        ssh_cmd = parse_ssh_option(self.config.ssh)[0].strip().lower()
         self.is_putty = ssh_cmd.endswith("plink") or ssh_cmd.endswith("plink.exe")
         self.is_paramiko = ssh_cmd.startswith("paramiko")
 
     def get_connection_modes(self):
-        modes = ["ssh"]
-        modes.append("ssh -> ssh")
+        modes = [MODE_SSH]
+        modes.append(MODE_NESTED_SSH)
         try:
             import ssl
             assert ssl
-            modes.append("ssl")
+            modes.append(MODE_SSL)
         except ImportError:
             pass
         #assume crypto is available
-        modes.append("tcp + aes")
-        modes.append("tcp")
-        modes.append("ws")
-        modes.append("wss")
+        try:
+            from xpra.net.crypto import MODES
+            for mode in MODES:
+                modes.append(f"tcp + aes-{mode.lower()}")
+        except ImportError:
+            pass
+        modes.append(MODE_TCP)
+        modes.append(MODE_WS)
+        modes.append(MODE_WSS)
         return modes
 
     def get_launcher_validation(self):
@@ -162,7 +184,7 @@ class ApplicationWindow:
         def validate_in_list(x, options):
             if x in options:
                 return None
-            return "must be in %s" % csv(options)
+            return "must be in "+csv(options)
         modes = self.get_connection_modes()
         return {"mode"              : lambda x : validate_in_list(x, modes)}
 
@@ -374,6 +396,15 @@ class ApplicationWindow:
         hbox.pack_start(al)
         vbox.pack_start(hbox)
 
+        #auto-connect
+        hbox = Gtk.HBox(False, 5)
+        self.autoconnect = Gtk.CheckButton("Auto-connect")
+        self.autoconnect.set_active(False)
+        al = Gtk.Alignment(xalign=0.5, yalign=0.5, xscale=0.0, yscale=0)
+        al.add(self.autoconnect)
+        hbox.pack_start(al)
+        vbox.pack_start(hbox)
+
         # Info Label
         self.info = Gtk.Label()
         self.info.set_line_wrap(True)
@@ -381,38 +412,6 @@ class ApplicationWindow:
         self.info.modify_fg(Gtk.StateType.NORMAL, red)
         vbox.pack_start(self.info)
 
-        hbox = Gtk.HBox(False, 0)
-        hbox.set_spacing(20)
-        self.advanced_options_check = Gtk.CheckButton("Advanced Options")
-        self.advanced_options_check.connect("toggled", self.advanced_options_toggled)
-        self.advanced_options_check.set_active(False)
-        al = Gtk.Alignment(xalign=0.5, yalign=0.5, xscale=0.0, yscale=0)
-        al.add(self.advanced_options_check)
-        hbox.pack_start(al)
-        vbox.pack_start(hbox)
-        self.advanced_box = Gtk.VBox()
-        vbox.pack_start(self.advanced_box)
-
-        # Encoding:
-        hbox = Gtk.HBox(False, 20)
-        hbox.set_spacing(20)
-        hbox.pack_start(Gtk.Label("Encoding: "))
-        self.encoding_combo = Gtk.MenuButton()
-        encodings = ["auto"]+[x for x in PREFERRED_ENCODING_ORDER]
-        server_encodings = encodings
-        es = make_encodingsmenu(self.get_current_encoding, self.set_new_encoding, encodings, server_encodings)
-        self.encoding_combo.set_popup(es)
-        hbox.pack_start(self.encoding_combo)
-        self.advanced_box.pack_start(hbox)
-        self.set_new_encoding(self.config.encoding)
-        # Sharing:
-        self.sharing = Gtk.CheckButton("Sharing")
-        self.sharing.set_active(bool(self.config.sharing))
-        self.sharing.set_tooltip_text("allow multiple concurrent users to connect")
-        al = Gtk.Alignment(xalign=0.5, yalign=0.5, xscale=0.0, yscale=0)
-        al.add(self.sharing)
-        self.advanced_box.pack_start(al)
-
         # Buttons:
         hbox = Gtk.HBox(False, 20)
         vbox.pack_start(hbox)
@@ -435,7 +434,6 @@ class ApplicationWindow:
         hbox.pack_start(self.connect_btn)
 
         vbox.show_all()
-        self.advanced_options_toggled()
         self.window.vbox = vbox
         self.window.add(vbox)
 
@@ -447,7 +445,7 @@ class ApplicationWindow:
         except Exception:
             pixbuf = None
         if not pixbuf:
-            pixbuf = get_icon_pixbuf("%s.png" % icon_name)
+            pixbuf = get_icon_pixbuf(f"{icon_name}.png")
             if pixbuf:
                 for size in (16, 32, 48):
                     scaled = pixbuf.scale_simple(size, size, GdkPixbuf.InterpType.BILINEAR)
@@ -475,8 +473,8 @@ class ApplicationWindow:
 
     def validate(self, *args):
         mode = self.mode_combo.get_active_text().lower()
-        ssh = mode=="ssh"
-        sshtossh = mode=="ssh -> ssh"
+        ssh = mode==MODE_SSH
+        sshtossh = mode==MODE_NESTED_SSH
         errs = []
         host = self.host_entry.get_text()
         errs.append((self.host_entry, not bool(host), "specify the host"))
@@ -537,7 +535,7 @@ class ApplicationWindow:
             self.set_widget_bg_color(w, e)
             if e:
                 err_text.append(text)
-        log("validate(%s) err_text=%s, errs=%s", args, err_text, errs)
+        log(f"validate({args}) err_text={err_text}, errs={errs}")
         self.set_info_text(csv(err_text))
         self.set_info_color(len(err_text)>0)
         self.connect_btn.set_sensitive(len(err_text)==0)
@@ -554,8 +552,8 @@ class ApplicationWindow:
 
     def mode_changed(self, *_args):
         mode = self.mode_combo.get_active_text().lower()
-        ssh = mode=="ssh"
-        sshtossh = mode=="ssh -> ssh"
+        ssh = mode==MODE_SSH
+        sshtossh = mode==MODE_NESTED_SSH
         if ssh or sshtossh:
             self.port_entry.set_tooltip_text("Display number (optional)")
             self.port_entry.set_text("")
@@ -583,7 +581,7 @@ class ApplicationWindow:
             self.password_entry.set_tooltip_text("Session Password (optional)")
             self.username_entry.set_tooltip_text("Session Username (optional)")
             if self.config.port>0:
-                self.port_entry.set_text("%s" % self.config.port)
+                self.port_entry.set_text(str(self.config.port))
         can_use_password = True
         sshpass = False
         if ssh or sshtossh:
@@ -623,44 +621,11 @@ class ApplicationWindow:
                 self.check_boxes_hbox.hide()
                 self.proxy_password_hbox.hide()
         self.validate()
-        if mode in ("ssl", "wss") or (mode=="ssh" and not WIN32):
+        if mode in (MODE_SSL, MODE_WSS) or (mode==MODE_SSH and not WIN32):
             self.nostrict_host_check.show()
         else:
             self.nostrict_host_check.hide()
 
-    def get_current_encoding(self):
-        return self.config.encoding
-
-    def set_new_encoding(self, e):
-        self.encoding_combo.set_label(e)
-        self.config.encoding = e
-
-    def get_selected_encoding(self, *_args):
-        if not self.encoding_combo:
-            return ""
-        index = get_active_item_index(self.encoding_combo)
-        return self.encoding_combo.get_popup().index_to_encoding.get(index)
-
-    def encoding_changed(self, *args):
-        encoding = self.get_selected_encoding()
-        uses_quality_option = encoding in ["jpeg", "webp", "h264", "auto"]
-        log("encoding_changed(%s) uses_quality_option(%s)=%s", args, encoding, uses_quality_option)
-        if uses_quality_option:
-            self.quality_combo.show()
-            self.quality_label.show()
-        else:
-            self.quality_combo.hide()
-            self.quality_label.hide()
-
-    def advanced_options_toggled(self, *_args):
-        if not self.advanced_box:
-            return
-        show_opts = self.advanced_options_check.get_active()
-        if show_opts:
-            self.advanced_box.show()
-        else:
-            self.advanced_box.hide()
-
     def reset_errors(self):
         self.set_sensitive(True)
         self.set_info_text("")
@@ -699,7 +664,7 @@ class ApplicationWindow:
         self.choose_pkey_file("Choose SSH private key", Gtk.FileChooserAction.OPEN, Gtk.STOCK_OPEN, do_choose)
 
     def connect_clicked(self, *args):
-        log("connect_clicked%s", args)
+        log(f"connect_clicked({args})")
         self.update_options_from_gui()
         self.do_connect()
 
@@ -715,7 +680,7 @@ class ApplicationWindow:
 
 
     def handle_exception(self, e):
-        log("handle_exception(%s)", e)
+        log(f"handle_exception({e})")
         t = str(e)
         log("handle_exception: %s", traceback.format_exc())
         if self.config.debug:
@@ -736,15 +701,17 @@ class ApplicationWindow:
         try:
             self.connect_builtin()
         except Exception as e:
-            log.error("cannot connect:", exc_info=True)
+            log.error("Error: cannot connect", exc_info=True)
             self.handle_exception(e)
 
     def connect_builtin(self):
         #cooked vars used by connect_to
-        params = {"type"    : self.config.mode}
-        self.config.sharing = self.sharing.get_active()
         username = self.config.username
-        if self.config.mode=="ssh" or self.config.mode=="ssh -> ssh":
+        params = {
+            "type"      : self.config.mode,
+            "username"  : username,
+            }
+        if self.config.mode in (MODE_SSH, MODE_NESTED_SSH):
             if self.config.socket_dir:
                 params["socket_dir"] = self.config.socket_dir
             params["remote_xpra"] = self.config.remote_xpra
@@ -771,18 +738,19 @@ class ApplicationWindow:
                     password = username[ppos+1:]
                     username = username[:ppos]
             if self.config.ssh_port and self.config.ssh_port!=22:
-                params["ssh-port"] = self.config.ssh_port
-            ssh_cmd = parse_ssh_string(self.config.ssh)
+                params["port"] = self.config.ssh_port
+            ssh_cmd = parse_ssh_option(self.config.ssh)
             ssh_cmd_0 = ssh_cmd[0].strip().lower()
             self.is_putty = ssh_cmd_0.endswith("plink") or ssh_cmd_0.endswith("plink.exe")
             self.is_paramiko = ssh_cmd_0 =="paramiko"
             full_ssh = ssh_cmd[:]
-            full_ssh += add_ssh_args(username, password, host, self.config.ssh_port, None, self.is_putty, self.is_paramiko)
+            full_ssh += add_ssh_args(username, password, host, self.config.ssh_port,
+                                     None, self.is_putty, self.is_paramiko)
             if username:
                 params["username"] = username
             if self.nostrict_host_check.get_active():
                 full_ssh += ["-o", "StrictHostKeyChecking=no"]
-            if params["type"] == "ssh -> ssh":
+            if params["type"] == MODE_NESTED_SSH:
                 params["type"] = "ssh"
                 params["proxy_host"] = self.config.proxy_host
                 params["proxy_port"] = self.config.proxy_port
@@ -796,17 +764,18 @@ class ApplicationWindow:
             params["local"] = is_local(self.config.host)
             params["full_ssh"] = full_ssh
             params["password"] = password
-            params["display_name"] = "ssh:%s:%s" % (self.config.host, self.config.port)
+            params["display_name"] = f"ssh://{self.config.host}:{self.config.port}"
         elif self.config.mode=="unix-domain":
-            params["display"] = ":%s" % self.config.port
-            params["display_name"] = "unix-domain:%s" % self.config.port
+            params["display"] = f":{self.config.port}"
+            params["display_name"] = f"unix-domain:{self.config.port}"
         else:
-            assert self.config.mode in ("tcp", "ssl", "ws", "wss"), "invalid / unsupported mode %s" % self.config.mode
+            if self.config.mode not in (MODE_TCP, MODE_SSL, MODE_WS, MODE_WSS):
+                raise ValueError(f"invalid / unsupported mode {self.config.mode}")
             params["host"] = self.config.host
             params["local"] = is_local(self.config.host)
             params["port"] = int(self.config.port)
-            params["display_name"] = "%s:%s:%s" % (self.config.mode, self.config.host, self.config.port)
-            if self.config.mode in ("ssl", "wss") and self.nostrict_host_check.get_active():
+            params["display_name"] = f"{self.config.mode}://{self.config.host}:{self.config.port}"
+            if self.config.mode in (MODE_SSL, MODE_WSS) and self.nostrict_host_check.get_active():
                 params["strict-host-check"] = False
 
         #print("connect_to(%s)" % params)
@@ -814,39 +783,44 @@ class ApplicationWindow:
         #or the config file may contain a username which is different from the default one
         #which is used for initializing the client during init,
         #so update the client now:
-        def raise_exception(*args):
-            raise Exception(*args)
         configure_env(self.config.env)
         configure_logging(self.config, "attach")
         configure_network(self.config)
+        self.start_client(params)
+
+    def start_client(self, display_desc):
+        def raise_exception(*args):
+            raise Exception(*args)
         self.client = make_client(raise_exception, self.config)
         self.client.show_progress(30, "client configuration")
         self.client.init(self.config)
         self.client.show_progress(40, "loading user interface")
         self.client.init_ui(self.config)
-        self.client.username = username
+        self.client.username = display_desc.get("username")
         def handshake_complete(*_args):
-            self.client.show_progress(100, "connection established")
+            self.client.show_progress(100, "Session connected")
         self.client.after_handshake(handshake_complete)
         self.set_info_text("Connecting...")
-        start_thread(self.do_connect_builtin, "connect", daemon=True, args=(params,))
+        start_thread(self.do_connect_builtin, "connect", daemon=True, args=(display_desc,))
 
     def ssh_failed(self, message):
-        log("ssh_failed(%s)", message)
+        log(f"ssh_failed({message})")
         if not self.current_error:
             self.current_error = message
             self.set_info_text(message)
             self.set_info_color(True)
 
     def do_connect_builtin(self, display_desc):
-        log("do_connect_builtin(%s)", display_desc)
+        log(f"do_connect_builtin({display_desc})")
         self.exit_code = None
         self.current_error = None
         self.set_info_text("Connecting.")
         self.set_sensitive(False)
         try:
-            log("calling %s%s", connect_to, (display_desc, repr_ellipsized(str(self.config)), self.set_info_text, self.ssh_failed))
-            conn = connect_to(display_desc, opts=self.config, debug_cb=self.set_info_text, ssh_fail_cb=self.ssh_failed)
+            log("calling %s%s", connect_to,
+                (display_desc, repr_ellipsized(str(self.config)), self.set_info_text, self.ssh_failed))
+            conn = connect_to(display_desc, opts=self.config,
+                              debug_cb=self.set_info_text, ssh_fail_cb=self.ssh_failed)
         except Exception as e:
             log("do_connect_builtin(%s) failed to connect", display_desc, exc_info=True)
             self.handle_exception(e)
@@ -854,12 +828,11 @@ class ApplicationWindow:
         log("connect_to(..)=%s, hiding launcher window, starting client", conn)
         GLib.idle_add(self.start_XpraClient, conn, display_desc)
 
-
     def start_XpraClient(self, conn, display_desc):
         try:
             self.do_start_XpraClient(conn, display_desc)
         except Exception as e:
-            log.error("failed to start client", exc_info=True)
+            log.error("Error: failed to start client", exc_info=True)
             self.handle_exception(e)
 
     def do_start_XpraClient(self, conn, display_desc):
@@ -868,7 +841,7 @@ class ApplicationWindow:
         self.client.display_desc = display_desc
         self.client.init_ui(self.config)
         self.client.setup_connection(conn)
-        self.set_info_text("Connection established")
+        self.set_info_text("Network connection established")
         log("start_XpraClient() client initialized")
 
         if self.config.password:
@@ -892,13 +865,25 @@ class ApplicationWindow:
                     self.set_sensitive(True)
                     GLib.idle_add(w.show)
 
+        def reconnect(exit_code):
+            log("reconnect(%s) config reconnect=%s",
+                EXIT_STR.get(exit_code, exit_code), self.config.reconnect)
+            if not self.config.reconnect or exit_code not in RETRY_EXIT_CODES:
+                return False
+            self.clean_client()
+            #give time for the main loop to run once after calling cleanup
+            GLib.timeout_add(100, self.start_client, display_desc)
+            return True
+
         def warn_and_quit_override(exit_code, warning):
             log("warn_and_quit_override(%s, %s)", exit_code, warning)
-            if self.exit_code is None:
-                self.exit_code = exit_code
             password_warning = warning.find("invalid password")>=0
             if password_warning:
                 self.password_warning()
+            elif reconnect(exit_code):
+                return
+            if self.exit_code is None:
+                self.exit_code = exit_code
             err = exit_code!=0 or password_warning
             if not self.current_error:
                 self.current_error = warning
@@ -908,6 +893,8 @@ class ApplicationWindow:
 
         def quit_override(exit_code):
             log("quit_override(%s)", exit_code)
+            if reconnect(exit_code):
+                return
             if self.exit_code is None:
                 self.exit_code = exit_code
             handle_client_quit(self.exit_code==0)
@@ -925,8 +912,8 @@ class ApplicationWindow:
             #no need to start a new main loop:
             self.client.gtk_main = noop
         try:
-            self.client.run()
-            log("client.run() returned")
+            r = self.client.run()
+            log("client.run() returned %s", r)
         except Exception as e:
             log.error("client error", exc_info=True)
             self.handle_exception(e)
@@ -975,6 +962,7 @@ class ApplicationWindow:
         self.config.port = pint(self.port_entry.get_text())
         self.config.username = self.username_entry.get_text()
         self.config.password = self.password_entry.get_text()
+        self.config.autoconnect = self.autoconnect.get_active()
 
         self.config.proxy_host = self.proxy_host_entry.get_text()
         self.config.proxy_port = pint(self.proxy_port_entry.get_text())
@@ -983,17 +971,17 @@ class ApplicationWindow:
         if self.is_putty:
             self.config.proxy_key = self.proxy_key_entry.get_text()
 
-        self.config.encoding = self.get_selected_encoding() or self.config.encoding
         mode_enc = self.mode_combo.get_active_text().lower()
-        if mode_enc.startswith("tcp"):
-            self.config.mode = "tcp"
+        if mode_enc.startswith(MODE_TCP):
+            self.config.mode = MODE_TCP
             if mode_enc.find("aes")>0:
-                self.config.encryption = "AES"
-        elif mode_enc in ("ssl", "ssh", "ws", "wss", "ssh -> ssh"):
+                self.config.encryption = "AES-"+mode_enc.split("aes-")[1].upper()
+        elif mode_enc in (MODE_SSL, MODE_SSH, MODE_WS, MODE_WSS, MODE_NESTED_SSH):
             self.config.mode = mode_enc
             self.config.encryption = ""
         log("update_options_from_gui() %s",
-            (self.config.username, self.config.password, self.config.mode, self.config.encryption, self.config.host, self.config.port, self.config.ssh_port, self.config.encoding))
+            (self.config.username, self.config.password, self.config.mode, self.config.encryption,
+             self.config.host, self.config.port, self.config.ssh_port))
 
     def update_gui_from_config(self):
         #mode:
@@ -1004,8 +992,6 @@ class ApplicationWindow:
                 active = i
                 break
         self.mode_combo.set_active(active)
-        if self.config.encoding and self.encoding_combo:
-            self.set_new_encoding(self.config.encoding)
         def get_port(vstr, default_port=""):
             try:
                 iport = int(vstr)
@@ -1015,7 +1001,7 @@ class ApplicationWindow:
                 pass
             return str(default_port)
         dport = DEFAULT_PORT
-        if mode in ("ssh", "ssh -> ssh"):
+        if mode in (MODE_SSH, MODE_NESTED_SSH):
             #not required, so don't specify one
             dport = ""
         self.port_entry.set_text(get_port(self.config.port, dport))
@@ -1033,7 +1019,6 @@ class ApplicationWindow:
         self.proxy_password_entry.set_text(proxy_password)
         if self.is_putty:
             self.proxy_key_entry.set_text(self.config.proxy_key)
-        self.sharing.set_active(bool(self.config.sharing))
         self.username_entry.set_text(username)
         self.password_entry.set_text(password)
         self.host_entry.set_text(self.config.host)
@@ -1056,7 +1041,7 @@ class ApplicationWindow:
         from xpra.scripts.parsing import parse_URL
         address, props = parse_URL(url)
         pa = address.split("://")
-        if pa[0] in ("tcp", "ssh", "ssl", "ws", "wss") and len(pa)>=2:
+        if pa[0] in (MODE_TCP, MODE_SSH, MODE_SSL, MODE_WS, MODE_WSS) and len(pa)>=2:
             props["mode"] = pa[0]
             host = pa[1]
             if host.find("@")>0:
@@ -1079,7 +1064,9 @@ class ApplicationWindow:
     def _apply_props(self, props):
         #we rely on "ssh_port" being defined on the config object
         #so try to load it from file, and define it if not present:
-        options = validate_config(props, extras_types=LAUNCHER_OPTION_TYPES, extras_validation=self.get_launcher_validation())
+        options = validate_config(props,
+                                  extras_types=LAUNCHER_OPTION_TYPES,
+                                  extras_validation=self.get_launcher_validation())
         for k,v in options.items():
             fn = k.replace("-", "_")
             setattr(self.config, fn, v)
@@ -1112,7 +1099,8 @@ class ApplicationWindow:
 #on some platforms like win32, we don't have stdout
 #and this is a GUI application, so show a dialog with the error instead
 def exception_dialog(title):
-    md = Gtk.MessageDialog(None, Gtk.DialogFlags.DESTROY_WITH_PARENT, Gtk.MessageType.INFO, Gtk.ButtonsType.CLOSE, title)
+    md = Gtk.MessageDialog(None, Gtk.DialogFlags.DESTROY_WITH_PARENT,
+                           Gtk.MessageType.INFO, Gtk.ButtonsType.CLOSE, title)
     md.format_secondary_text(traceback.format_exc())
     md.show_all()
     def close_dialog(*_args):
@@ -1134,10 +1122,9 @@ def do_main(argv):
     from xpra.os_util import SIGNAMES
     from xpra.scripts.main import InitExit, InitInfo
     from xpra.platform.gui import init as gui_init, ready as gui_ready
+    from xpra.gtk_common.gtk_util import init_display_source
 
-    if POSIX and not OSX:
-        from xpra.x11.gtk_x11.gdk_display_source import init_gdk_display_source
-        init_gdk_display_source()
+    init_display_source()
 
     gui_init()
     try:
@@ -1169,7 +1156,7 @@ def do_main(argv):
                 client.cleanup()
             else:
                 Gtk.main_quit()
-            GLib.timeout_add(1000, app.set_info_text, "got signal %s" % SIGNAMES.get(signum, signum))
+            GLib.timeout_add(1000, app.set_info_text, "got signal " + SIGNAMES.get(signum, signum))
             GLib.timeout_add(1000, app.set_info_color, True)
         register_os_signals(handle_signal, "Client Launcher")
         has_file = len(args) == 1
