@@ -1,21 +1,21 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 from math import sqrt
-from time import sleep
+from time import sleep, monotonic
 
 from xpra.server.source.stub_source_mixin import StubSourceMixin
-from xpra.server.window.batch_config import DamageBatchConfig
+from xpra.server.window import batch_config
 from xpra.server.server_core import ClientException
 from xpra.codecs.video_helper import getVideoHelper
 from xpra.codecs.codec_constants import video_spec
-from xpra.net.compression import compressed_wrapper, Compressed, use_lz4, use_lzo
-from xpra.os_util import monotonic_time, strtobytes
+from xpra.net.compression import use
+from xpra.os_util import bytestostr
 from xpra.server.background_worker import add_work_item
 from xpra.util import csv, typedict, envint
 from xpra.log import Logger
@@ -25,35 +25,30 @@ proxylog = Logger("proxy")
 statslog = Logger("stats")
 
 MIN_PIXEL_RECALCULATE = envint("XPRA_MIN_PIXEL_RECALCULATE", 2000)
+DEFAULT_VREFRESH = envint("XPRA_DEFAULT_VREFRESH", 100)
 
 
-"""
-Store information about the client's support for encodings.
-Runs the encode thread.
-"""
 class EncodingsMixin(StubSourceMixin):
+    """
+    Store information about the client's support for encodings.
+    Runs the encode thread.
+    """
 
-    def __init__(self):
-        self.server_core_encodings = []
-        self.server_encodings = []
-        self.default_encoding = None
-        self.scaling_control = None
-
-        self.default_quality = 40       #default encoding quality for lossy encodings
-        self.default_min_quality = 10   #default minimum encoding quality
-        self.default_speed = 40         #encoding speed (only used by x264)
-        self.default_min_speed = 10     #default minimum encoding speed
+    @classmethod
+    def is_needed(cls, caps : typedict) -> bool:
+        return bool(caps.strtupleget("encodings"))
 
+    def init_state(self):
         #contains default values, some of which may be supplied by the client:
-        self.default_batch_config = DamageBatchConfig()
+        self.default_batch_config = batch_config.DamageBatchConfig()
         self.global_batch_config = self.default_batch_config.clone()      #global batch config
 
-        self.vrefresh = -1
         self.supports_transparency = False
         self.encoding = None                        #the default encoding for all windows
         self.encodings = ()                         #all the encodings supported by the client
         self.core_encodings = ()
-        self.window_icon_encodings = ["premult_argb32"]
+        self.encodings_packet = False               #supports delayed encodings initialization?
+        self.window_icon_encodings = []
         self.rgb_formats = ("RGB",)
         self.encoding_options = typedict()
         self.icons_encoding_options = typedict()
@@ -61,8 +56,7 @@ class EncodingsMixin(StubSourceMixin):
         self.auto_refresh_delay = 0
 
         self.zlib = True
-        self.lz4 = use_lz4
-        self.lzo = use_lzo
+        self.lz4 = use("lz4")
 
         #for managing the recalculate_delays work:
         self.calculate_window_pixels = {}
@@ -74,6 +68,7 @@ class EncodingsMixin(StubSourceMixin):
         #new encoders, so we must make a deep copy to preserve the original
         #which may be used by other clients (other ServerSource instances)
         self.video_helper = getVideoHelper().clone()
+        self.cuda_device_context = None
 
 
     def init_from(self, _protocol, server):
@@ -86,8 +81,14 @@ class EncodingsMixin(StubSourceMixin):
         self.default_speed          = server.default_speed
         self.default_min_speed      = server.default_min_speed
 
+    def reinit_encodings(self, server):
+        self.server_core_encodings  = server.core_encodings
+        self.server_encodings       = server.encodings
+
     def cleanup(self):
         self.cancel_recalculate_timer()
+        if self.cuda_device_context:
+            self.queue_encode((False, self.free_cuda_device_context))
         #Warning: this mixin must come AFTER the window mixin!
         #to make sure that it is safe to add the end of queue marker:
         #(all window sources will have stopped queuing data)
@@ -95,29 +96,29 @@ class EncodingsMixin(StubSourceMixin):
         #this should be a noop since we inherit an initialized helper:
         self.video_helper.cleanup()
 
+    def free_cuda_device_context(self):
+        cdd = self.cuda_device_context
+        if cdd:
+            self.cuda_device_context = None
+            cdd.free()
 
-    def get_caps(self):
+    def all_window_sources(self):
+        #we can't assume that the window mixin is loaded:
+        window_sources = getattr(self, "window_sources", {})
+        return tuple(window_sources.values())
+
+
+    def get_caps(self) -> dict:
         caps = {}
-        if self.wants_encodings and self.encoding:
+        if "encodings" in self.wants and self.encoding:
             caps["encoding"] = self.encoding
-        if self.wants_features:
+        if "features" in self.wants:
             caps.update({
                 "auto_refresh_delay"   : self.auto_refresh_delay,
                 })
         return caps
 
 
-    def compressed_wrapper(self, datatype, data, min_saving=128):
-        if self.zlib or self.lz4 or self.lzo:
-            cw = compressed_wrapper(datatype, data, zlib=self.zlib, lz4=self.lz4, lzo=self.lzo, can_inline=False)
-            if len(cw)+min_saving<=len(data):
-                #the compressed version is smaller, use it:
-                return cw
-            #skip compressed version: fall through
-        #we can't compress, so at least avoid warnings in the protocol layer:
-        return Compressed(datatype, data, can_inline=True)
-
-
     def recalculate_delays(self):
         """ calls update_averages() on ServerSource.statistics (GlobalStatistics)
             and WindowSource.statistics (WindowPerformanceStatistics) for each window id in calculate_window_ids,
@@ -126,16 +127,19 @@ class EncodingsMixin(StubSourceMixin):
         self.calculate_timer = 0
         if self.is_closed():
             return
-        now = monotonic_time()
+        now = monotonic()
         self.calculate_last_time = now
         p = self.protocol
-        if not p:
+        if not p or p.is_closed():
             return
         conn = p._conn
         if not conn:
             return
-        self.statistics.bytes_sent.append((now, conn.output_bytecount))
-        self.statistics.update_averages()
+        #we can't assume that 'self' is a full ClientConnection object:
+        stats = getattr(self, "statistics", None)
+        if stats:
+            stats.bytes_sent.append((now, conn.output_bytecount))
+            stats.update_averages()
         self.update_bandwidth_limits()
         wids = tuple(self.calculate_window_ids)  #make a copy so we don't clobber new wids
         focus = self.get_focus()
@@ -147,10 +151,7 @@ class EncodingsMixin(StubSourceMixin):
         for wid in wids:
             #this is safe because we only add to this set from other threads:
             self.calculate_window_ids.remove(wid)
-            try:
-                del self.calculate_window_pixels[wid]
-            except:
-                pass
+            self.calculate_window_pixels.pop(wid, None)
             ws = self.window_sources.get(wid)
             if ws is None:
                 continue
@@ -160,7 +161,7 @@ class EncodingsMixin(StubSourceMixin):
                                          len(fullscreen_wids)>0 and wid not in fullscreen_wids,
                                          len(maximized_wids)>0 and wid not in maximized_wids)
                 ws.reconfigure()
-            except:
+            except Exception:
                 log.error("error on window %s", wid, exc_info=True)
             if self.is_closed():
                 return
@@ -205,7 +206,7 @@ class EncodingsMixin(StubSourceMixin):
         if self.calculate_timer:
             #already due
             return
-        delta = monotonic_time() - self.calculate_last_time
+        delta = monotonic() - self.calculate_last_time
         RECALCULATE_DELAY = 1.0           #1s
         if delta>RECALCULATE_DELAY:
             add_work_item(self.recalculate_delays)
@@ -220,7 +221,7 @@ class EncodingsMixin(StubSourceMixin):
             self.source_remove(ct)
 
 
-    def parse_client_caps(self, c):
+    def parse_client_caps(self, c : typedict):
         #batch options:
         def batch_value(prop, default, minv=None, maxv=None):
             assert default is not None
@@ -250,33 +251,39 @@ class EncodingsMixin(StubSourceMixin):
 
         #general features:
         self.zlib = c.boolget("zlib", True)
-        self.lz4 = c.boolget("lz4", False) and use_lz4
-        self.lzo = c.boolget("lzo", False) and use_lzo
+        self.lz4 = c.boolget("lz4", False) and use("lz4")
+        self.brotli = c.boolget("brotli", False) and use("brotli")
+        log("compressors: zlib=%s, lz4=%s, brotli=%s",
+            self.zlib, self.lz4, self.brotli)
 
+        delay = batch_config.START_DELAY
+        dbc = self.default_batch_config
+        dbc.always      = bool(batch_value("always", dbc.always))
+        dbc.min_delay   = batch_value("min_delay", dbc.min_delay, 0, 1000)
+        dbc.max_delay   = batch_value("max_delay", dbc.max_delay, 1, 15000)
+        dbc.max_events  = batch_value("max_events", dbc.max_events)
+        dbc.max_pixels  = batch_value("max_pixels", dbc.max_pixels)
+        dbc.time_unit   = batch_value("time_unit", dbc.time_unit, 1)
+        dbc.delay       = batch_value("delay", delay, dbc.min_delay)
+        log("default batch config: %s", dbc)
         self.vrefresh = c.intget("vrefresh", -1)
 
-        default_min_delay = max(DamageBatchConfig.MIN_DELAY, 1000//(self.vrefresh or 60))
-        dbc = self.default_batch_config
-        dbc.always      = bool(batch_value("always", DamageBatchConfig.ALWAYS))
-        dbc.min_delay   = batch_value("min_delay", default_min_delay, 0, 1000)
-        dbc.max_delay   = batch_value("max_delay", DamageBatchConfig.MAX_DELAY, 1, 15000)
-        dbc.max_events  = batch_value("max_events", DamageBatchConfig.MAX_EVENTS)
-        dbc.max_pixels  = batch_value("max_pixels", DamageBatchConfig.MAX_PIXELS)
-        dbc.time_unit   = batch_value("time_unit", DamageBatchConfig.TIME_UNIT, 1)
-        dbc.delay       = batch_value("delay", DamageBatchConfig.START_DELAY, 0)
-        log("default batch config: %s", dbc)
-
         #encodings:
-        self.encodings = c.strlistget("encodings")
-        self.core_encodings = c.strlistget("encodings.core", self.encodings)
-        if self.send_windows and not self.core_encodings:
+        self.encodings_packet = c.boolget("encodings.packet", False)
+        self.encodings = c.strtupleget("encodings")
+        self.core_encodings = c.strtupleget("encodings.core", self.encodings)
+        log("encodings=%s, core_encodings=%s", self.encodings, self.core_encodings)
+        #we can't assume that the window mixin is loaded,
+        #or that the ui_client flag exists:
+        send_ui = getattr(self, "ui_client", True) and getattr(self, "send_windows", True)
+        if send_ui and not self.core_encodings:
             raise ClientException("client failed to specify any supported encodings")
-        if "png" in self.core_encodings:
-            self.window_icon_encodings.append("png")
-        self.window_icon_encodings = c.strlistget("encodings.window-icon", ["premult_argb32"])
-        self.rgb_formats = c.strlistget("encodings.rgb_formats", ["RGB"])
+        self.window_icon_encodings = c.strtupleget("encodings.window-icon", ())
+        #try both spellings for older versions:
+        for x in ("encodings", "encoding",):
+            self.rgb_formats = c.strtupleget(x+".rgb_formats", self.rgb_formats)
         #skip all other encoding related settings if we don't send pixels:
-        if not self.send_windows:
+        if not send_ui:
             log("windows/pixels forwarding is disabled for this client")
         else:
             self.parse_encoding_caps(c)
@@ -300,18 +307,18 @@ class EncodingsMixin(StubSourceMixin):
         #2: standardized encoding options:
         for k in c.keys():
             #yaml gives us str..
-            k = strtobytes(k)
-            if k.startswith(b"theme.") or k.startswith(b"encoding.icons."):
-                self.icons_encoding_options[k.replace(b"encoding.icons.", b"").replace(b"theme.", b"")] = c.get(k)
-            elif k.startswith(b"encoding."):
-                stripped_k = k[len(b"encoding."):]
-                if stripped_k in (b"transparency",
-                                  b"rgb_zlib", b"rgb_lz4", b"rgb_lzo",
-                                  b"video_scaling"):
+            k = bytestostr(k)
+            if k.startswith("theme.") or k.startswith("encoding.icons."):
+                self.icons_encoding_options[k.replace("encoding.icons.", "").replace("theme.", "")] = c.get(k)
+            elif k.startswith("encoding."):
+                stripped_k = k[len("encoding."):]
+                if stripped_k in ("transparency",
+                                  "rgb_zlib", "rgb_lz4",
+                                  ):
                     v = c.boolget(k)
-                elif stripped_k in (b"initial_quality", b"initial_speed",
-                                    b"min-quality", b"quality",
-                                    b"min-speed", b"speed"):
+                elif stripped_k in ("initial_quality", "initial_speed",
+                                    "min-quality", "quality",
+                                    "min-speed", "speed"):
                     v = c.intget(k)
                 else:
                     v = c.get(k)
@@ -326,7 +333,7 @@ class EncodingsMixin(StubSourceMixin):
             #enabling video proxy:
             try:
                 self.parse_proxy_video()
-            except:
+            except Exception:
                 proxylog.error("failed to parse proxy video", exc_info=True)
 
         sc = self.encoding_options.get("scaling.control", self.scaling_control)
@@ -348,23 +355,46 @@ class EncodingsMixin(StubSourceMixin):
             self.default_encoding_options["min-speed"] = ms
         log("default encoding options: %s", self.default_encoding_options)
         self.auto_refresh_delay = c.intget("auto_refresh_delay", 0)
-        #check for mmap:
-        if getattr(self, "mmap_size", 0)==0:
-            others = tuple(x for x in self.core_encodings
-                           if x in self.server_core_encodings and x!=self.encoding)
-            if self.encoding=="auto":
-                s = "automatic picture encoding enabled"
-            else:
-                s = "using %s as primary encoding" % self.encoding
-            if others:
-                log.info(" %s, also available:", s)
-                log.info("  %s", csv(others))
-            else:
-                log.warn(" %s", s)
-                log.warn("  no other encodings are available!")
+
+        #are we going to need a cuda context?
+        if getattr(self, "mmap_enabled", False):
+            #not with mmap!
+            return
+        common_encodings = tuple(x for x in self.encodings if x in self.server_encodings)
+        from xpra.codecs.loader import has_codec
+        if (has_codec("nvenc") and "h264" in self.core_encodings or "h265" in self.core_encodings) \
+        or ("jpeg" in common_encodings and has_codec("enc_nvjpeg")):
+            cudalog = Logger("cuda")
+            try:
+                from xpra.codecs.cuda_common.cuda_context import get_device_context
+                self.cuda_device_context = get_device_context(self.encoding_options)
+                cudalog("cuda_device_context=%s", self.cuda_device_context)
+            except Exception as e:
+                cudalog("failed to get a cuda device context using encoding options %s",
+                    self.encoding_options, exc_info=True)
+                cudalog.error("Error: failed to allocate a CUDA context:")
+                cudalog.error(" %s", e)
+                cudalog.error(" NVJPEG and NVENC will not be available")
+
+    def print_encoding_info(self):
+        log("print_encoding_info() core-encodings=%s, server-core-encodings=%s",
+            self.core_encodings, self.server_core_encodings)
+        others = tuple(x for x in self.core_encodings
+                       if x in self.server_core_encodings and x!=self.encoding)
+        if self.encoding=="auto":
+            s = "automatic picture encoding enabled"
+        else:
+            s = "using %s as primary encoding" % self.encoding
+        if others:
+            log.info(" %s, also available:", s)
+            log.info("  %s", csv(others))
+        else:
+            log.warn(" %s", s)
+            log.warn("  no other encodings are available!")
 
     def parse_proxy_video(self):
-        from xpra.codecs.enc_proxy.encoder import Encoder
+        self.wait_for_threaded_init()
+        from xpra.codecs.enc_proxy.encoder import Encoder  #pylint: disable=import-outside-toplevel
         proxy_video_encodings = self.encoding_options.get("proxy.video.encodings")
         proxylog("parse_proxy_video() proxy.video.encodings=%s", proxy_video_encodings)
         for encoding, colorspace_specs in proxy_video_encodings.items():
@@ -373,7 +403,7 @@ class EncodingsMixin(StubSourceMixin):
                     #make a new spec based on spec_props:
                     spec_prop = typedict(spec_prop)
                     input_colorspace = spec_prop.strget("input_colorspace")
-                    output_colorspaces = spec_prop.strlistget("output_colorspaces")
+                    output_colorspaces = spec_prop.strtupleget("output_colorspaces")
                     if not input_colorspace or not output_colorspaces:
                         log.warn("Warning: invalid proxy video encoding '%s':", encoding)
                         log.warn(" missing colorspace attributes")
@@ -397,22 +427,20 @@ class EncodingsMixin(StubSourceMixin):
     # Functions used by the server to request something
     # (window events, stats, user requests, etc)
     #
-    def set_auto_refresh_delay(self, delay, window_ids):
+    def set_auto_refresh_delay(self, delay : int, window_ids):
         if window_ids is not None:
             wss = (self.window_sources.get(wid) for wid in window_ids)
         else:
-            wss = self.window_sources.values()
+            wss = self.all_window_sources()
         for ws in wss:
             if ws is not None:
                 ws.set_auto_refresh_delay(delay)
 
-    def set_encoding(self, encoding, window_ids, strict=False):
+    def set_encoding(self, encoding : str, window_ids, strict=False):
         """ Changes the encoder for the given 'window_ids',
             or for all windows if 'window_ids' is None.
         """
         log("set_encoding(%s, %s, %s)", encoding, window_ids, strict)
-        if not self.ui_client:
-            return
         if encoding and encoding!="auto":
             #old clients (v0.9.x and earlier) only supported 'rgb24' as 'rgb' mode:
             if encoding=="rgb24":
@@ -422,17 +450,21 @@ class EncodingsMixin(StubSourceMixin):
                 log.warn(" but it only supports: %s" % csv(self.encodings))
             if encoding not in self.server_encodings:
                 log.error("Error: encoding %s is not supported by this server", encoding)
+                log.error(" server encodings: %s", csv(self.server_encodings))
                 encoding = None
         if not encoding:
             encoding = "auto"
         if window_ids is not None:
             wss = [self.window_sources.get(wid) for wid in window_ids]
         else:
-            wss = self.window_sources.values()
+            wss = self.all_window_sources()
         #if we're updating all the windows, reset global stats too:
-        if set(wss).issuperset(self.window_sources.values()):
+        if set(wss).issuperset(self.all_window_sources()):
             log("resetting global stats")
-            self.statistics.reset()
+            #we can't assume that 'self' is a full ClientConnection object:
+            stats = getattr(self, "statistics", None)
+            if stats:
+                stats.reset()
             self.global_batch_config = self.default_batch_config.clone()
         for ws in wss:
             if ws is not None:
@@ -441,18 +473,13 @@ class EncodingsMixin(StubSourceMixin):
             self.encoding = encoding
 
 
-    def get_info(self):
+    def get_info(self) -> dict:
         info = {
                 "auto_refresh"      : self.auto_refresh_delay,
                 "lz4"               : self.lz4,
-                "lzo"               : self.lzo,
-                "vertical-refresh"  : self.vrefresh,
                 }
         ieo = dict(self.icons_encoding_options)
-        try:
-            del ieo["default.icons"]
-        except KeyError:
-            pass
+        ieo.pop("default.icons", None)
         #encoding:
         info.update({
                      "encodings"        : {
@@ -471,39 +498,34 @@ class EncodingsMixin(StubSourceMixin):
         return info
 
 
-    def set_min_quality(self, min_quality):
-        for ws in tuple(self.window_sources.values()):
+    def set_min_quality(self, min_quality : int):
+        for ws in tuple(self.all_window_sources()):
             ws.set_min_quality(min_quality)
 
-    def set_quality(self, quality):
-        for ws in tuple(self.window_sources.values()):
+    def set_max_quality(self, max_quality : int):
+        for ws in tuple(self.all_window_sources()):
+            ws.set_max_quality(max_quality)
+
+    def set_quality(self, quality : int):
+        for ws in tuple(self.all_window_sources()):
             ws.set_quality(quality)
 
-    def set_min_speed(self, min_speed):
-        for ws in tuple(self.window_sources.values()):
+    def set_min_speed(self, min_speed : int):
+        for ws in tuple(self.all_window_sources()):
             ws.set_min_speed(min_speed)
 
-    def set_speed(self, speed):
-        for ws in tuple(self.window_sources.values()):
+    def set_max_speed(self, max_speed : int):
+        for ws in tuple(self.all_window_sources()):
+            ws.set_max_speed(max_speed)
+
+    def set_speed(self, speed : int):
+        for ws in tuple(self.all_window_sources()):
             ws.set_speed(speed)
 
 
-    def update_batch(self, wid, window, batch_props):
-        ws = self.window_sources.get(wid)
-        if ws:
-            if "reset" in batch_props:
-                ws.batch_config = self.make_batch_config(wid, window)
-            for x in ("always", "locked"):
-                if x in batch_props:
-                    setattr(ws.batch_config, x, batch_props.boolget(x))
-            for x in ("min_delay", "max_delay", "timeout_delay", "delay"):
-                if x in batch_props:
-                    setattr(ws.batch_config, x, batch_props.intget(x))
-            log("batch config updated for window %s: %s", wid, ws.batch_config)
-
-    def make_batch_config(self, wid, window):
-        batch_config = self.default_batch_config.clone()
-        batch_config.wid = wid
+    def make_batch_config(self, wid : int, window):
+        config = self.default_batch_config.clone()
+        config.wid = wid
         #scale initial delay based on window size
         #(the global value is normalized to 1MPixel)
         #but use sqrt to smooth things and prevent excesses
@@ -513,7 +535,7 @@ class EncodingsMixin(StubSourceMixin):
         w, h = window.get_dimensions()
         if dpm>=0:
             ratio = sqrt(1000000.0 / (w*h))
-            batch_config.delay = max(batch_config.min_delay, min(batch_config.max_delay, int(dpm * sqrt(ratio))))
+            config.delay = max(config.min_delay, min(config.max_delay, int(dpm * sqrt(ratio))))
         log("make_batch_config(%i, %s) global delay per megapixel=%i, new window delay for %ix%i=%s",
-                 wid, window, dpm, w, h, batch_config.delay)
-        return batch_config
+                 wid, window, dpm, w, h, config.delay)
+        return config
